from shoulder import utils
from shoulder.base import Landmark

import ruptures
import numpy as np
import scipy.signal
import skspatial.objects
import plotly.graph_objects as go
from functools import cached_property
import matplotlib.pyplot as plt


class DeepGroove(Landmark):
    def __init__(self, obb):
        self._mesh_oriented_uobb = obb.mesh
        self._transform_uobb = obb.transform
        self._obb_cutoff_pcts = obb.cutoff_pcts
        self._points_ct = None
        self._points = None
        self._axis_ct = None
        self._axis = None

    def axis(self, cutoff_pcts=[0.35, 0.85], slice_num=35, interp_num=250):
        if self._axis is None:
            proximal_cutoff, distal_cutoff = self._surgical_neck_cutoff_zs(*cutoff_pcts)
            # slice_num  must use odd soo add 1 if even
            if (slice_num % 2) == 0:
                slice_num += 1

            # find z interval to calculate bicipital groove upon
            # self.cutoff_pcts.sort()
            # get length of the bone
            # z_max = np.max(self._mesh_oriented_uobb.bounds[:, -1])
            # z_min = np.min(self._mesh_oriented_uobb.bounds[:, -1])
            # z_length = abs(z_max) + abs(z_min)

            # # find distance that the cutoff percentages are at
            # # cutoff_pcts.sort()  # ensure bottom slice pct is first
            # distal_cutoff = self.cutoff_pcts[0] * z_length + z_min
            # proximal_cutoff = self.cutoff_pcts[1] * z_length + z_min

            zs = np.linspace(distal_cutoff, proximal_cutoff, num=slice_num).flatten()

            # preallocate variables
            xy = np.zeros((interp_num, 2, slice_num))
            polar = np.zeros((interp_num, 2, slice_num))
            weights = np.zeros((interp_num, 2, slice_num))
            to_3Ds = np.zeros((4, 4, slice_num))

            for i, z in enumerate(zs):
                # grab the polygon of the slice
                origin = [0, 0, z]
                normal = [0, 0, 1]
                path = self._mesh_oriented_uobb.section(
                    plane_origin=origin, plane_normal=normal
                )
                slice, to_3D = path.to_planar(normal=normal)
                # keep only largest polygon
                big_poly = slice.polygons_closed[
                    np.argmax([p.area for p in slice.polygons_closed])
                ]

                # resample cartesion coordinates to create evenly spaced points
                _pts = np.asarray(big_poly.exterior.xy).T
                _pts = _resample_polygon(_pts, interp_num)

                # convert to polar and ensure even degree spacing
                _pol = _cart2pol(_pts)

                # if a cavity is present do not count that as a weight
                theta_diff = (
                    np.diff(_pol[:, 0], prepend=-10) < 0
                )  # prepend -10 so first difference is positive
                cav = _true_propogate(theta_diff)  # all cavities are True
                cav = np.array(cav, dtype=np.int32)  # make all true 1
                cav = cav ^ (
                    cav & 1 == cav
                )  # flip all 0s to 1s, since we want to preserve everythin but cavities
                cav_weight = np.c_[cav, cav]

                # log data
                xy[:, :, i] = _pts
                polar[:, :, i] = _pol
                weights[:, :, i] = cav_weight
                to_3Ds[:, :, i] = to_3D

            # make each radial slice stationary
            polar_0 = polar.copy()
            polar_0[:, 1, :] = np.apply_along_axis(
                lambda x: x - np.mean(x), axis=0, arr=polar[:, 1, :]
            )

            # calculate weighted mean across slices where cavities have a weight of 0
            polar_avg_0 = np.average(polar_0, axis=2, weights=weights)
            deg = np.rad2deg(polar_avg_0[:, 0])
            radius = polar_avg_0[:, 1]
            # weights create jagged edges
            radius = scipy.signal.savgol_filter(radius, 10, 1)

            # calulate derivatives
            dd_radius = _derivative_smooth_ends(radius, 2, 10, 2)

            peaks, _prop = scipy.signal.find_peaks(
                dd_radius, height=0, distance=interp_num / 360 * 25
            )
            peaks = peaks[
                np.argpartition(_prop["peak_heights"], -3)[-3:]
            ]  # top 3 largest

            # find the peaks that are not near the furthest point
            # the furthest point is on the articular surface so any peaks neighbouring there
            # would not be the biciptal groove
            peaks.sort()
            deg_rmax = deg[np.argmax(radius)]
            deg_peaks = deg[peaks]
            filt_vals = np.r_[deg_peaks, deg_rmax]
            deg_shft = np.r_[deg[peaks[0] :], deg[: peaks[0]], deg[peaks[0]]]
            filt = [x for x in deg_shft if x in filt_vals]
            non_bg_peaks = (
                filt[filt.index(deg_rmax) - 1],
                filt[filt.index(deg_rmax) + 1],
            )
            bg_peak = list(set(deg_peaks) - set(non_bg_peaks))[0]

            # get local minima by specifying serach window for
            # search up to 15 degrees away on each side
            deg_variance = int(round(360 / interp_num) * 15)
            bg_xyz = np.zeros((1, 3, len(zs)))
            for i, z in enumerate(zs):
                # print(polar_0)
                bg_idx_near = _find_nearest_idx(
                    polar_0[:, 0, i].flatten(), np.deg2rad(bg_peak)
                )
                # sometimes the degree variance will be higher than than the index bg is found at
                # when this occurs the indexing will start with a negative numebr causing it to fail
                # basically a wrap around problem

                if deg_variance > bg_idx_near:
                    bg_range = np.concatenate(
                        (
                            polar_0[(bg_idx_near - deg_variance) :, :, i],
                            polar_0[: (bg_idx_near + deg_variance), :, i],
                        ),
                        axis=0,
                    )
                else:
                    bg_range = polar_0[
                        (bg_idx_near - deg_variance) : (bg_idx_near + deg_variance),
                        :,
                        i,
                    ]
                bg_local_i = np.argmin(bg_range[:, 1])

                # transform back to radial coordinates
                bg_i = bg_local_i + (bg_idx_near - deg_variance)  # put back in context
                _bg_xy = _pol2cart(polar[bg_i, :, i].reshape(1, 2))

                # i think to_3D is perhaps fully broken, doesn't seem to work
                bg_xyz[:, :, i] = utils.transform_pts(np.c_[_bg_xy, 0], to_3Ds[:, :, i])

            bg_xyz = bg_xyz.transpose(2, 1, 0).reshape(-1, 3)

            # transform back
            bg_xyz = utils.transform_pts(
                bg_xyz, utils.inv_transform(self._transform_uobb)
            )

            # construct an estimate of the bicipital groove axis from the bg_xyz pts
            line_ends = _fit_line(bg_xyz)

            self._axis_ct = line_ends
            self._axis = line_ends
            self._points_ct = bg_xyz
            self._points = bg_xyz

        return self._axis

    def transform_landmark(self, transform) -> None:
        if self._axis is not None:
            self._points = utils.transform_pts(self._points_ct, transform)
            self._axis = utils.transform_pts(self._axis_ct, transform)

    def _graph_obj(self):
        if self._points is None:
            return None

        else:
            plot = go.Scatter3d(
                x=self._points[:, 0],
                y=self._points[:, 1],
                z=self._points[:, 2],
                name="Bicipital Groove",
            )
            return plot

    def _surgical_neck_cutoff_zs(self, bottom_pct=0.35, top_pct=0.85):
        """given cutoff perccentages with 0 being the surgical neck and 1 being the
        top of the head return the z coordaintes
        """
        # this basically calcuates where the surgical neck is
        z_max = np.max(self._mesh_oriented_uobb.bounds[:, -1])
        z_min = np.min(self._mesh_oriented_uobb.bounds[:, -1])
        z_length = abs(z_max) + abs(z_min)

        z_low_pct = self._obb_cutoff_pcts[0]
        z_high_pct = self._obb_cutoff_pcts[1]
        distal_cutoff = z_low_pct * z_length + z_min
        proximal_cutoff = z_high_pct * z_length + z_min
        # print(distal_cutoff)

        z_intervals = np.linspace(distal_cutoff, 0.99 * z_max, 100)

        z_area = np.zeros(len(z_intervals))
        for i, z in enumerate(z_intervals):
            slice = self._mesh_oriented_uobb.section(
                plane_origin=[0, 0, z], plane_normal=[0, 0, 1]
            )
            slice, to_3d = slice.to_planar()
            # big_poly = slice.polygons_closed[
            #     np.argmax([p.area for p in slice.polygons_closed])
            # ]
            z_area[i,] = slice.area

        algo = ruptures.KernelCPD(kernel="rbf")
        algo.fit(z_area)
        bkp = algo.predict(n_bkps=1)

        surgical_neck_z = z_intervals[bkp[0]]
        surgical_neck_top_head = z_max - surgical_neck_z
        bottom = surgical_neck_z + (surgical_neck_top_head * bottom_pct)
        top = surgical_neck_z + (surgical_neck_top_head * top_pct)

        # interval on which to calcaulte bicipital groove
        return [bottom, top]

    def _apply_cutoff_pcts(self, cutoff_pcts):
        z_max = np.max(self._mesh_oriented_uobb.bounds[:, -1])
        z_min = np.min(self._mesh_oriented_uobb.bounds[:, -1])
        z_length = abs(z_max) + abs(z_min)

        # find distance that the cutoff percentages are at
        # cutoff_pcts.sort()  # ensure bottom slice pct is first
        distal_cutoff = self.cutoff_pcts[0] * z_length + z_min
        proximal_cutoff = self.cutoff_pcts[1] * z_length + z_min

        return [distal_cutoff, proximal_cutoff]


def _find_nearest_idx(array, value):
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (
        idx == len(array) or np.abs(value - array[idx - 1]) < np.abs(value - array[idx])
    ):
        return idx - 1
    else:
        return idx


def _reorder_by_theta(arr):
    """reorder the array to start at the most negative theta.
    only works when it is an ordered array from a path object"""

    re_arr = np.r_[arr[np.argmin(arr[:, 0]) :], arr[: np.argmin(arr[:, 0])]]
    # there is an error that can occur when the most positive number which should be at
    # the end has a negative theta
    if re_arr[-1, 0] < 0:
        re_arr[-1, 0] = np.deg2rad(179.99)

    return re_arr


def _cart2pol(arr: np.ndarray) -> np.ndarray:
    """convert from cartesian coordinates to radial

    Args:
        arr (np.ndarray): cartesian coordiantes

    Returns:
        np.ndarray: radial coordinates
    """
    x = arr[:, 0]
    y = arr[:, 1]
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    r_arr = np.c_[theta, r]
    # r_arr = r_arr[r_arr[:,1].argsort()] # sort by theta angle
    r_arr = _reorder_by_theta(r_arr)
    return r_arr


def _pol2cart(arr):
    r = arr[:, 1]
    theta = arr[:, 0]
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return np.c_[x, y]


def _resample_polygon(xy: np.ndarray, n_points: int = 100) -> np.ndarray:
    """interpolate between points in array to ensure even spacing between all points

    Args:
        xy (np.ndarray): array with columns for x and y coordinates arranged by order in
            which they occur while tracing along edge of polygon.
        n_points (int, optional): number of evenly spaced points to return. Defaults to 100.

    Returns:
        np.ndarray: evenly spaced points
    """
    # Cumulative Euclidean distance between successive polygon points.
    # This will be the "x" for interpolation
    d = np.cumsum(np.r_[0, np.sqrt((np.diff(xy, axis=0) ** 2).sum(axis=1))])

    # get linearly spaced points along the cumulative Euclidean distance
    d_sampled = np.linspace(0, d.max(), n_points)

    # interpolate x and y coordinates
    xy_interp = np.c_[
        np.interp(d_sampled, d, xy[:, 0]),
        np.interp(d_sampled, d, xy[:, 1]),
    ]

    return xy_interp


def _derivative_smooth(arr, dd_order, window, smooth_order, _depth=0):
    """calculates nth order derivatives while smoothing in-between each step

    Args:
        arr (np.array): 1D array to calculate the derivative of
        dd_order (int): order of derivate to calculate
        window (int): interval upon which the smoothing occurs
        smooth_order (int): order of function which smoothing uses
        _depth (int, optional): . Defaults to 0.

    Returns:
        np.array: derivative of array of order n
    """
    if _depth == dd_order:
        return scipy.signal.savgol_filter(arr, window, smooth_order)
    arr = scipy.signal.savgol_filter(np.gradient(arr), window, smooth_order)
    return _derivative_smooth(arr, dd_order, window, smooth_order, _depth + 1)


def _derivative_smooth_ends(arr, dd_order, window, smooth_order, _depth=0):
    """calculates nth order derivatives while smoothing at the end

    Args:
        arr (np.array): 1D array to calculate the derivative of
        dd_order (int): order of derivate to calculate
        window (int): interval upon which the smoothing occurs
        smooth_order (int): order of function which smoothing uses
        _depth (int, optional): . Defaults to 0.

    Returns:
        np.array: derivative of array of order n
    """
    if _depth == dd_order:  # initial pass
        return scipy.signal.savgol_filter(arr, window, smooth_order)
    elif _depth == 0:  # final pass
        arr = scipy.signal.savgol_filter(arr, window, smooth_order)
    arr = np.gradient(arr)
    return _derivative_smooth(arr, dd_order, window, smooth_order, _depth + 1)


def _true_propogate(arr):
    """for each true interval double the size starting at same position"""

    def true_interval(x):
        """find interval where true bools  occur"""
        z = np.concatenate(([False], x, [False]))

        start = np.flatnonzero(~z[:-1] & z[1:])
        end = np.flatnonzero(z[:-1] & ~z[1:])

        return np.column_stack((start, end))

    b_i = true_interval(arr)

    a = arr.copy()

    for i in b_i:
        start = i[0]
        end = i[1]
        length = end - start
        # make sure that there is enoguh space
        if length > len(a[end : end + length]):
            continue
        else:
            a[end : end + length] = np.repeat(True, length)

    return a


def _fit_line(bg_xyz):
    x, y, z = bg_xyz.T
    z_dist = np.max(z) - np.min(z)
    line_fit = skspatial.objects.Line.best_fit(bg_xyz)
    ends = np.array(
        [
            line_fit.point + (line_fit.direction * (z_dist / 2)),
            line_fit.point - (line_fit.direction * (z_dist / 2)),
        ]
    )

    return ends



from functools import cached_property, lru_cache
import warnings
from pathlib import Path
import trimesh
import circle_fit
import numpy as np
import scipy.signal
from abc import ABC, abstractmethod


import matplotlib.pyplot as plt


class MeshLoader:
    def __init__(self, stl_file) -> None:
        if not isinstance(stl_file, Path):
            stl_file = Path(stl_file)

        self.file = stl_file
        self.name = stl_file.stem

    @cached_property
    def mesh_ct(self):
        m = trimesh.load_mesh(str(self.file))
        if not m.is_watertight:
            warnings.warn(f"{self.name} is not watertight!")
        return m


class Obb(ABC, MeshLoader):
    @property
    @abstractmethod
    def mesh(self) -> trimesh.Trimesh:
        """mesh with transform applied"""

    @property
    @abstractmethod
    def transform(self) -> np.ndarray:
        """oriented bounding box transformation matrix (4x4)"""

    @property
    @abstractmethod
    def cutoff_pcts(self) -> list:
        """cutoff of bounding box for uneven cut of proximal humerus"""

    @cached_property
    @abstractmethod
    def _obb(self) -> list:
        """calculates the oriented bouding box returns _mesh, _transform, _cutoff_pcts(optional)"""


class FullObb(Obb):
    def __init__(self, stl_file):
        super().__init__(stl_file)

    @property
    def mesh(self) -> trimesh.Trimesh:
        return self._obb[0]

    @property
    def transform(self) -> np.ndarray:
        return self._obb[1]

    @property
    def cutoff_pcts(self):
        return [0.5, 0.8]

    @cached_property
    def _obb(self):
        """rotates the humerus so the humeral head faces up (+ y-axis)

        Args:
            mesh (trimesh.mesh): mesh to rotate up

        Returns:
            mesh: rotated mesh
            ct_T: transform back to CT space
        """

        """ The center of volume is now at (0,0,0) with the y axis of the CSYS being the long axis of the humerus.
        The z being left-right and the x being up-down when viewed along the y-axis (humeral-axis)
        Whether the humeral head lies in +y space or -y space is unkown. The approach to discover which end is which
        is to take a slice on each end and see which shape is more circular. The more circular end is obviously the
        humeral head.
        """

        # apply oriented bounding box
        _mesh = self.mesh_ct.copy()
        _transform_obb = _mesh.apply_obb()  # modify in place returns transform

        # Get z bounds of box
        z_limits = (_mesh.bounds[0][-1], _mesh.bounds[1][-1])

        # look at slice shape on each end
        humeral_end = 0  # y-coordinate default value
        residu_init = np.inf  # residual of circle fit default value
        for z_limit in z_limits:
            # make the slice
            # move 5% inwards on the half, so 2.5% of total length
            z_slice = 0.95 * z_limit
            slice = _mesh.section(plane_origin=[0, 0, z_slice], plane_normal=[0, 0, 1])
            # returns the 2d view at plane and the transformation back to 3d space for each point
            slice, to_3d = slice.to_planar()

            # pull out the points along the shapes edge
            _, _, _, residu = circle_fit.least_squares_circle(np.array(slice.vertices))

            # 1st pass, less than inf record, 2nd pass if less than 1st
            if residu < residu_init:
                residu_init = residu
                humeral_end = z_limit

        # if the y-coordinate of the humeral head is in negative space then
        # we are looking to see if a flip was performed and if it was needed
        # humeral_end is a set containing (y-coordinate, residual from circle fit)
        if humeral_end < 0:
            # print("flipped")
            # flip was reversed so update the ct_transform to refelct that
            transform_flip = np.array(
                [[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
            )
            _mesh.apply_transform(transform_flip)
        else:
            # print("not flipped")
            transform_flip = np.array(
                [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
            )

        # add in flip that perhaps occured
        _transform = np.matmul(transform_flip, _transform_obb)
        return _mesh, _transform


class ProxObb(Obb):
    def __init__(self, stl_file):
        super().__init__(stl_file)

    @property
    def mesh(self) -> trimesh.Trimesh:
        return self._obb[0]

    @property
    def transform(self) -> np.ndarray:
        return self._obb[1]

    @property
    def cutoff_pcts(self) -> list:
        return self._obb[2]

    @cached_property
    def _obb(self):
        """to determine which side is the humeral head and which side is the cut shaft is a
        simple comparison of area at each end. The cut is not always clean and is sometimes angled
        which presents some difficulties.
        To overcome this get the area of 100 points along the z axis of the obb. If the area becomes smaller
        towards the end remove if from consideration."""

        def consecutive(arr):
            return max(np.split(arr, (np.where(np.diff(arr) != 1)[0] + 1)), key=len)

        # apply oriented bounding box
        _mesh = self.mesh_ct.copy()
        _transform_obb = _mesh.apply_obb()  # modify in place returns transform

        # Get z bounds of box
        z_limits = (_mesh.bounds[0][-1], _mesh.bounds[1][-1])

        # find largest area along z axis
        inset_factor = 0.99  # percent shrink z of first slice
        # evenly space z intervals
        num_zs = 100
        z_intervals = np.linspace(
            z_limits[0] * inset_factor, z_limits[1] * inset_factor, num_zs
        ).flatten()
        z_area = []
        for z in z_intervals:
            slice = _mesh.section(plane_origin=[0, 0, z], plane_normal=[0, 0, 1])
            slice, to_3d = slice.to_planar()
            z_area.append(slice.area)

        # the middle of humeral head has the largest area for proximal humerus
        humeral_head_z = z_intervals[np.argmax(z_area)]

        # flip so the humeral head is up
        if humeral_head_z < 0:
            transform_flip = np.array(
                [[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
            )
            _mesh.apply_transform(transform_flip)
            # reverse the order now that it has been flipped
            z_area = z_area[::-1]
        else:
            transform_flip = np.array(
                [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
            )
        # add in flip that perhaps occured
        _transform = np.matmul(transform_flip, _transform_obb)

        # calculate the difference between one slice and the next but smoothed
        grad_z_area = np.gradient(scipy.signal.savgol_filter(z_area, 3, 1))

        # keep gradients smaller than a diff of 5, these must be the canal as it changes little in area
        # this will also remove the improperly cut portion
        canal_zs = consecutive(np.where(grad_z_area < 10)[0])

        # cutoff percentages for when canal needs to be found
        cutoff_pcts = [canal_zs[0] / num_zs, canal_zs[-1] / num_zs]

        return _mesh, _transform, cutoff_pcts



# prevent circular import errors
from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from . import base

# import other packages
import numpy as np
import stl
import plotly.graph_objects as go


class Plot:
    def __init__(
        self,
        bone: base.Bone,
        opacity=1.0,
    ):
        self.opacity = opacity
        self.stl_name = bone.stl_file.name
        self.transform = bone.transform
        self.stl_mesh = stl.mesh.Mesh.from_file(bone.stl_file)
        self.stl_mesh.transform(self.transform)
        self._landmarks_graph_obj = bone._list_landmarks_graph_obj()

    def stl2mesh3d(self, stl_mesh):
        # stl_mesh is read by nympy-stl from a stl file; it is  an array of faces/triangles (i.e. three 3d points)
        # this function extracts the unique vertices and the lists I, J, K to define a Plotly mesh3d
        p, q, r = stl_mesh.vectors.shape  # (p, 3, 3)
        # the array stl_mesh.vectors.reshape(p*q, r) can contain multiple copies of the same vertex;
        # extract unique vertices from all mesh triangles
        vertices, ixr = np.unique(
            stl_mesh.vectors.reshape(p * q, r), return_inverse=True, axis=0
        )
        I = np.take(ixr, [3 * k for k in range(p)])
        J = np.take(ixr, [3 * k + 1 for k in range(p)])
        K = np.take(ixr, [3 * k + 2 for k in range(p)])
        return vertices, I, J, K

    @property
    def figure(self):
        vertices, I, J, K = self.stl2mesh3d(self.stl_mesh)
        x, y, z = vertices.T

        # add stl
        fig = go.Figure(
            data=[
                go.Mesh3d(
                    x=x,
                    y=y,
                    z=z,
                    i=I,
                    j=J,
                    k=K,
                    opacity=self.opacity,
                    color="grey",
                    lighting=dict(
                        ambient=0.18,
                        diffuse=0.8,
                        fresnel=0.1,
                        specular=1.2,
                        roughness=0.05,
                        facenormalsepsilon=1e-15,
                        vertexnormalsepsilon=1e-15,
                    ),
                    lightposition=dict(x=1000, y=1000, z=-1000),
                    flatshading=False,
                )
            ]
        )
        fig.update_layout(
            title=self.stl_name, scene_aspectmode="data"
        )  # plotly defualts into focing 3d plots to be distorted into cubes, this prevents that

        for lgo in self._landmarks_graph_obj:
            # if more than one graph object per landmark class
            if isinstance(lgo, list):
                for l in lgo:
                    fig.add_trace(l)
            else:
                fig.add_trace(lgo)

        return fig



from shoulder import utils
from shoulder.humerus import mesh
from shoulder.base import Landmark

import plotly.graph_objects as go
import numpy as np
from skspatial.objects import Line, Points


class Canal(Landmark):
    def __init__(self, obb: mesh.Obb):
        """Calculates the centerline of the humeral canal"""
        self._mesh_oriented_uobb = obb.mesh
        self._transform_uobb = obb.transform
        self.cutoff_pcts = obb.cutoff_pcts
        self._points_ct = None
        self._points = None
        self._axis_ct = None
        self._axis = None

    def axis(self, cutoff_pcts: list = None, num_slices: int = 50) -> np.ndarray:
        """calculates the centerline in region of humerus

        Args:
            cutoff_pcts (list): cutoff for where centerline is to be fit between i.e [0.2,0.8] -> middle 60% of the bone

            num_slices (int): number of slices to generate between cutoff points for which centroids will be calculated

        Returns:
            canal_pts_ct: 2x3 matrix of xyz points at ends of centerline
        """

        def axial_centroids(msh_o, cutoff_pcts, num_centroids):
            """Slices bone along long axis between the specified cutoff points and returns
            the centroids along the length

            Args:
                mesh (trimesh.mesh): trimesh mesh object to find centroids along length
                cutoff_pcts (list): list of two cutoff percentages i.e [0.2,0.8] would remove the upper 20% and lower 20%
                num_centroids (int): number of slices beween cutoff points to calculate centroids for

            Returns:
                centroids (np.array): array of xyz points for centroids along length
                cutoff_length (float): length between the cutoff percentages on the bone
            """

            # get length of the bone
            z_max = np.max(msh_o.bounds[:, -1])
            z_min = np.min(msh_o.bounds[:, -1])
            z_length = abs(z_max) + abs(z_min)

            # find distance that the cutoff percentages are at
            cutoff_pcts.sort()  # ensure bottom slice pct is first
            distal_cutoff = cutoff_pcts[0] * z_length + z_min
            proximal_cutoff = cutoff_pcts[1] * z_length + z_min
            # length between cutoff pts
            cutoff_length = abs(proximal_cutoff - distal_cutoff)

            # spacing of cuts
            cuts = np.linspace(
                distal_cutoff, proximal_cutoff, num=num_centroids
            ).flatten()

            centroids = []  # record data
            for cut in cuts:
                slice = msh_o.section(plane_origin=[0, 0, cut], plane_normal=[0, 0, 1])
                centroids.append(np.array(slice.centroid).reshape(1, 3))

            centroids = np.concatenate(centroids, axis=0)

            return centroids, cutoff_length

        if self._axis is None:
            if cutoff_pcts is not None:
                self.cutoff_pcts = cutoff_pcts

            # slice it !
            centroids, cutoff_length = axial_centroids(
                self._mesh_oriented_uobb, self.cutoff_pcts, num_slices
            )

            # transform back then record the centroids
            centroids_ct = utils.transform_pts(
                centroids, utils.inv_transform(self._transform_uobb)
            )
            self._points = centroids_ct
            self._points_ct = centroids_ct

            # calculate centerline
            canal_fit = Line.best_fit(Points(centroids))
            canal_direction = canal_fit.direction
            canal_mdpt = canal_fit.point

            # ensure that the vector is pointed proximally
            if canal_fit.direction[-1] < 0:
                canal_direction = canal_direction * -1

            # repersent centerline as two points at the extents of the cutoff
            canal_prox = canal_mdpt + (canal_direction * (cutoff_length / 2))
            canal_dstl = canal_mdpt - (canal_direction * (cutoff_length / 2))
            canal_pts = np.array([canal_prox, canal_dstl])
            canal_pts_ct = utils.transform_pts(
                canal_pts, utils.inv_transform(self._transform_uobb)
            )

            self._axis_ct = canal_pts_ct
            self._axis = canal_pts_ct  # will be transformed later
        return self._axis

    # get_transform method only needed for the first axis of a csys i.e. the independent axis
    def get_transform(self) -> np.ndarray:
        """Get transform from CT csys to a csys with z-axis as the canal.

        Args:
            canal_axis (np.ndarray): 2x3 matrix of xyz points at ends of centerline
            _transform_uobb (np.ndarray): transform matrix from the CT csys to the OBB csys

        Returns:
            np.ndarray: 4x4 transform matrix from the CT csys to the canal csys

        Take the canal as the z axis x axis of the OBB csys and project it onto a plane
        orthgonal to the canal axis, this will make the x axis othogonal to the canal axis.
        Then take the cross product to find the last axis. This creates a transform from the
        canal csys to the ct csys but we would like the opposite so invert it before returning the transform
        """
        # canal axis
        z_hat = utils.unit_vector(self._axis[0], self._axis[1])
        # grab x axis from OBB csys
        x_hat = self._transform_uobb[:3, :1].flatten()

        # project to be orthogonal
        x_hat -= z_hat * np.dot(x_hat, z_hat) / np.dot(z_hat, z_hat)
        x_hat /= np.linalg.norm(x_hat)

        # find last axis
        y_hat = np.cross(z_hat, x_hat)
        y_hat /= np.linalg.norm(y_hat)

        # assemble
        pos = np.average(self._axis, axis=0)
        transform = np.c_[x_hat, y_hat, z_hat, pos]
        transform = np.r_[transform, np.array([0, 0, 0, 1]).reshape(1, 4)]

        # return a transform that goes form CT_csys -> Canal_csys
        transform = utils.inv_transform(transform)

        return transform

    def transform_landmark(self, transform) -> None:
        self._axis = utils.transform_pts(self._axis_ct, transform)
        self._points = utils.transform_pts(self._points_ct, transform)

    def _graph_obj(self):
        if self._points is None:
            return None
        else:
            plot = go.Scatter3d(
                x=self._points[:, 0],
                y=self._points[:, 1],
                z=self._points[:, 2],
                name="Canal Axis",
            )
            return plot



from shoulder.humerus import mesh
from shoulder.humerus import canal
from shoulder.base import Landmark
from shoulder import utils

import plotly.graph_objects as go
import shapely
import numpy as np
import itertools


class TransEpicondylar(Landmark):
    def __init__(self, obb: mesh.FullObb):
        self._mesh_oriented = obb.mesh
        self._transform = obb.transform
        self._axis_ct = None
        self._axis = None

    def axis(self, num_slices: int = 50):
        def medial_lateral_dist_multislice(msh_o, num_slices: int):
            """slices along distal humerus and computes the medial lateral distance with a rotated bounding box

            Args:
                msh_o (trimesh.mesh): rotated trimesh mesh object
                num_slices (int): number of slices to make between 10% and 0%

            """
            # get length of the tranformed bone
            total_length = np.sum(abs(msh_o.bounds[:, -1]))  # entire length of bone
            pos_length = msh_o.bounds[
                msh_o.bounds[:, -1] >= 0, -1
            ]  # bone in positive space, bone before the centerline midpoint

            proximal_cutoff = -0.8 * total_length + pos_length
            distal_cutoff = -0.99 * total_length + pos_length

            # spacing of cuts
            cuts = np.linspace(proximal_cutoff, distal_cutoff, num=num_slices).flatten()

            dist = []

            for cut in cuts:
                try:
                    path = msh_o.section(
                        plane_origin=[0, 0, cut], plane_normal=[0, 0, 1]
                    )
                    slice, to_3d = path.to_planar()
                except:
                    break

                # get shapely object from path
                polygon = slice.polygons_closed[0]

                # create rotated bounding box
                majr_dist = utils.major_axis_dist(
                    polygon.minimum_rotated_rectangle
                )  # maximize this distance
                dist.append(majr_dist)
            idx_max_dist = dist.index(max(dist))
            max_dist_cut = cuts[idx_max_dist]

            return max_dist_cut

        if self._axis is None:
            # find z distance where medial lateral distance is longest
            z_dist = medial_lateral_dist_multislice(self._mesh_oriented, num_slices)

            # slice at location of max medial-lateral distance
            path = self._mesh_oriented.section(
                plane_normal=[0, 0, 1], plane_origin=[0, 0, z_dist]
            )
            slice, to_3d = path.to_planar()

            # get shapely object from path
            polygon = slice.polygons_closed[0]

            # create rotated bounding box
            bound = polygon.minimum_rotated_rectangle
            bound_angle = utils.azimuth(bound)

            # cut ends off at edge of bounding box that align with major axis
            bound_scale = shapely.affinity.rotate(bound, bound_angle)
            bound_scale = shapely.affinity.scale(bound_scale, xfact=1.5, yfact=0.999)
            bound_scale = shapely.affinity.rotate(bound_scale, -bound_angle)
            ends = polygon.difference(bound_scale)

            # now we have the most medial and lateral points
            # sometimes one of the end sections can be split in two leaving more than 2 total ends
            if len(list(ends.geoms)) > 2:
                ab_dists = []
                # iterate through all distance combos
                for a, b in itertools.combinations(list(ends.geoms), 2):
                    ab_dists.append(
                        [
                            a,
                            b,
                            utils._dist(
                                np.array(a.centroid.xy).flatten(),
                                np.array(b.centroid.xy).flatten(),
                            ),
                        ]
                    )  # [obj,obj,distance]
                end_geoms = list(
                    np.array(ab_dists)[np.argmax(np.array(ab_dists)[:, 2]), :2]
                )  # find location of max distance return shapely objs
                end_pts = np.array(
                    [end_geoms[0].centroid.xy, end_geoms[1].centroid.xy]
                ).reshape(2, 2)
            else:
                end_pts = np.array(
                    [ends.geoms[0].centroid.xy, ends.geoms[1].centroid.xy]
                ).reshape(2, 2)

            # add z distance back ins
            end_pts = utils.z_zero_col(end_pts)

            # transform back
            end_pts = utils.transform_pts(end_pts, to_3d)
            end_pts_ct = utils.transform_pts(
                end_pts, utils.inv_transform(self._transform)
            )

            self._axis_ct = end_pts_ct
            self._axis = end_pts_ct
        return self._axis

    def transform_landmark(self, transform) -> None:
        self._axis = utils.transform_pts(self._axis_ct, transform)

    def _graph_obj(self):
        if self._axis is None:
            return None
        else:
            plot = go.Scatter3d(
                x=self._axis[:, 0],
                y=self._axis[:, 1],
                z=self._axis[:, 2],
                name="Transverse Epicondylar Axis",
            )
            return plot



""" angles

This module contains functions for the matrix math that determines neck shaft angle and retroversion.
"""

from shoulder import utils

import trimesh
import skspatial
import numpy as np




def neck_shaft(anp_normal, _transform, _transform_arp):
    """measures angle in degrees between anatomic neck plane and canal axis

    Args:
        anp_normal (array): array of vector for anatomic neck plane normal
        _transform (array): transform to move to bone specific coordiante system        
        _transform_arp (array): transform view to perpedicular to anp_normal and remove retroversion angle  

    Returns:
        array: anatomic neck shaft angle
    """

    _transform_ns = np.matmul(_transform_arp, _transform)
    _anp_normal = utils.transform_pts(anp_normal, _transform_ns)
    _anp_normal = skspatial.objects.Line.from_points(
        _anp_normal[0, :].reshape(
            3,
        ),
        _anp_normal[1, :].reshape(
            3,
        ),
    ).direction
    # out of plane for neck shaft plane is now the y direction
    _anp_normal[1] = 0

    _, neck_shaft_angle = trimesh.geometry.align_vectors(
        _anp_normal, np.array([0, 0, 1]), return_angle=True
    )

    neck_shaft_angle = 180 - np.rad2deg(neck_shaft_angle)
    # print(neck_shaft_angle)
    return neck_shaft_angle


def retroversion(anp_normal, side, _transform):
    """measured between transepicondylar axis and the articular margin plane normal projected onto the axial(z) plane

    Args:
        trep_axis (array): transepiconylar axis
        anp_normal (array): articular margin plane normal vector

    Returns:
        _transform_arp (array): transfrom to rotate so med-lat axis is parallel to articular margin plane normal projected in axial
        rv_angle (float): angle in degrees of the anatomic retroversion between the transepicondylar axis and the anatomic neck plane
    """
    # project the anp_normal onto the z-plane
    _anp_normal = utils.transform_pts(anp_normal, _transform)

    # _anp_normal = skspatial.objects.Line(_anp_normal[0, :], _anp_normal[1, :]).direction
    _anp_normal = skspatial.objects.Line.from_points(
        _anp_normal[0, :].reshape(
            3,
        ),
        _anp_normal[1, :].reshape(
            3,
        ),
    ).direction

    # ignore z direction
    _anp_normal[-1] = 0

    if side == "left":
        _transform_arp, rv_angle = trimesh.geometry.align_vectors(
            _anp_normal, np.array([-1, 0, 0]), return_angle=True
        )
        # posterior should always be negative, currently is positive
        transform_lr = np.array(
            [[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
        )
        _transform_arp = np.matmul(transform_lr, _transform_arp)
    else:
        _transform_arp, rv_angle = trimesh.geometry.align_vectors(
            _anp_normal, np.array([-1, 0, 0]), return_angle=True
        )

    # depending on clockwise or counterclockwise rotation correct it.
    rv_angle = np.rad2deg(rv_angle)
    if rv_angle > 90:
        rv_angle = 360 - rv_angle

    # transform_amptrep = utils.inv_transform(transform_amptrep)
    # print(rv_angle)
    return _transform_arp, rv_angle



import numpy as np
import scipy.stats
import scipy.spatial


def write_iges_line(line, filepath):
    x, y, z = line[0]
    x1, y1, z1 = line[1]

    # i known  this string looks jank but it has to be this way for the whitespace to be printed correct
    s = """                                                                        S0000001
1H,,1H;,8Hpart.mco,10Hglobal.tmp,21HMedcad by Materialise,              G0000001
24HMedical CAD Modelling sw,32,38,8,308,16,8Hpart.med,1,2,2HMM,3,0.5,   G0000002
13H220408.155753,9.9999999E-09,1000,2HPN,14HMaterialise nv,1,0;         G0000003
     110       1       0       1       0       0       0       000000000D0000001
     110       0       0       1       0                    LINE       0D0000002\n"""
    s1 = f"110,{x},{y},{z},{x1},{y1},{z1};"
    s1 = s1.ljust(71) + "1P0000001\n"
    s2 = "S      1G      3D      2P      1                                        T0000001"
    iges = s + s1 + s2

    with open(filepath, "w") as f:
        f.write(iges)


def z_score_filter(arr, idx, threshold):
    i = arr - np.median(arr, axis=0)
    i = np.abs(scipy.stats.zscore(i)[:, idx]) < threshold

    return arr[i]


# BOUNDING BOX MATH #
#####################
def azimuth(mrr):
    """azimuth of minimum_rotated_rectangle"""
    bbox = np.asarray(mrr.exterior.xy).T
    axis1 = _dist(bbox[0], bbox[3])
    axis2 = _dist(bbox[0], bbox[1])
    # print(axis1,axis2)
    if axis1 <= axis2:
        az = _azimuth(bbox[0], bbox[1])
    else:
        az = _azimuth(bbox[0], bbox[3])

    return az


def _azimuth(point1, point2):
    """azimuth between 2 points (interval 0 - 180)"""
    import numpy as np

    angle = np.arctan2(point2[0] - point1[0], point2[1] - point1[1])
    return np.degrees(angle) if angle > 0 else np.degrees(angle) + 180


def _dist(a, b):
    """distance between points"""
    import math

    return math.hypot(b[0] - a[0], b[1] - a[1])


def major_axis(mrr):
    bbox = np.array((mrr.exterior.xy)).T
    axis1 = _dist(bbox[0], bbox[3])
    axis2 = _dist(bbox[0], bbox[1])

    if axis1 > axis2:
        coords = np.stack(
            [
                np.mean(np.stack([bbox[0], bbox[1]], axis=0), axis=0),
                np.mean(np.stack([bbox[2], bbox[3]], axis=0), axis=0),
            ],
            axis=0,
        )
    else:
        coords = np.stack(
            [
                np.mean(np.stack([bbox[0], bbox[3]], axis=0), axis=0),
                np.mean(np.stack([bbox[1], bbox[2]], axis=0), axis=0),
            ],
            axis=0,
        )

    return coords


def major_axis_dist(mrr):
    bbox = np.array((mrr.exterior.xy)).T
    axis1 = _dist(bbox[0], bbox[3])
    axis2 = _dist(bbox[0], bbox[1])

    if axis1 > axis2:
        return axis1
    else:
        return axis2


def minor_axis(mrr):
    bbox = np.array((mrr.exterior.xy)).T
    axis1 = _dist(bbox[0], bbox[3])
    axis2 = _dist(bbox[0], bbox[1])

    if axis1 < axis2:
        coords = np.stack(
            [
                np.mean(np.stack([bbox[0], bbox[1]], axis=0), axis=0),
                np.mean(np.stack([bbox[2], bbox[3]], axis=0), axis=0),
            ],
            axis=0,
        )
    else:
        coords = np.stack(
            [
                np.mean(np.stack([bbox[0], bbox[3]], axis=0), axis=0),
                np.mean(np.stack([bbox[1], bbox[2]], axis=0), axis=0),
            ],
            axis=0,
        )

    return coords


def minor_axis_dist(mrr):
    bbox = np.array((mrr.exterior.xy)).T
    axis1 = _dist(bbox[0], bbox[3])
    axis2 = _dist(bbox[0], bbox[1])

    if axis1 > axis2:
        return axis2
    else:
        return axis1


def closest_pt(pt, pts, return_other_pts=False):
    """find closest point to the array of points"""
    kdtree = scipy.spatial.cKDTree(pts)
    d, i = kdtree.query(pt)  # returns distance and loction in index of closest point
    if return_other_pts:
        return [
            pts[i],
            np.delete(pts, i, axis=0),
        ]  # delete doesn't modify the orginal array
    else:
        return pts[i]


# MATRIX MATH #
###############
def rot_matrix_3d(vec1, vec2):
    """Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (
        vec2 / np.linalg.norm(vec2)
    ).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s**2))
    return rotation_matrix


def z_zero_col(xy):
    return np.c_[xy, np.zeros(len(xy))]


def transform_pts(pts, transform):
    """Applies a transform to a set of xyz points

    Args:
        pts (np.array [nx3]): points to transform
        transform (np.array [4x4]): transformation matrix

    Returns:
        pts_transform(np.array [nx3]): transformed points
    """
    pts = np.c_[pts, np.ones(len(pts))].T  # add column of ones then transpose -> 4xn
    pts = np.matmul(transform, pts)
    pts = pts.T  # transpose back
    pts_transform = np.delete(
        pts, 3, axis=1
    )  # remove added ones now that transform is complete
    return pts_transform


def inv_transform(transform):
    """inverses a transformation matrix

    Args:
        transform (np.array [4x4]): transformation matrix

    Returns:
        transform (np.array [4x4]): inverse transformation matrix
    """
    # make translation its own 4x4 matrix
    translate = transform[:3, -1]  # pull out 1x3 translation column matrix
    translate = np.c_[
        np.identity(3), translate
    ]  # add 3x3 identity matrix -> 3x4 matrix
    translate = np.r_[
        translate, np.array([[0, 0, 0, 1]])
    ]  # add row of 0 0 0 1 -> 4x4 matrix

    # make rotation its own 4x4 matrix
    rotate = transform[:, :-1]  # take everythin but last column -> 4x3 matrix
    rotate = np.c_[
        rotate, np.array([[0], [0], [0], [1]])
    ]  # replace with null translation -> 4x4 matrix

    # multiply the inverted matrices
    transform = np.matmul(
        np.linalg.inv(rotate), np.linalg.inv(translate)
    )  # R^-1 x T^-1

    return transform


def unit_vector(p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
    vec = p1 - p2
    unit_vec = vec / np.linalg.norm(vec)

    return unit_vec



from . import utils
from .humerus import mesh
from .humerus import canal
from .humerus import epicondyle
from .humerus import anatomic_neck
from .humerus import bicipital_groove
from .base import Bone

from abc import ABC, abstractmethod
from pathlib import Path
import plotly.graph_objects as go
import numpy as np

# ignore warnings
# This is bad practice but one of the libraries has an issue that generates a nonsense warning
import warnings

warnings.filterwarnings("ignore")


class Humerus(Bone):
    def __init__(self, stl_file):
        self.stl_file = stl_file
        self.transform = np.identity(4)
        msh = mesh.FullObb(stl_file)
        self._mesh = msh

        self.canal = canal.Canal(msh)
        self.trans_epiconylar = epicondyle.TransEpicondylar(msh)
        self.anatomic_neck = anatomic_neck.AnatomicNeck(msh, self.trans_epiconylar)
        self.bicipital_groove = bicipital_groove.DeepGroove(msh)

    def apply_csys_canal_transepiconylar(self) -> np.ndarray:
        self.transform = construct_csys(self.canal._axis, self.trans_epiconylar._axis)
        self._update_landmark_data(self.transform)
        return self.transform


class ProximalHumerus(Bone):
    def __init__(self, stl_file):
        self.stl_file = stl_file
        self.transform = np.identity(4)

        msh = mesh.ProxObb(stl_file)
        self._mesh = msh

        self.canal = canal.Canal(msh)
        self.bicipital_groove = bicipital_groove.DeepGroove(msh)

    def apply_csys_canal_articular(self, articular) -> np.ndarray:
        self.transform = construct_csys(self.canal._axis, articular)
        self._update_landmark_data(self.transform)
        return self.transform


def construct_csys(vec_z, vec_y):
    # define center and two axes
    pos = np.average(vec_z, axis=0)
    pos = pos.flatten()
    z_hat = utils.unit_vector(vec_z[0], vec_z[1])
    x_hat = utils.unit_vector(vec_y[0], vec_y[1])

    # calculate remaing axis
    y_hat = np.cross(x_hat, z_hat)
    y_hat /= np.linalg.norm(y_hat)

    # transepicondylar axis is not quite perpendicular so do it again
    # this is but a temporary fix, maybe switchinf back the transepi to
    # being dependent on the canal would be wise
    x_hat = np.cross(y_hat, z_hat)
    x_hat /= np.linalg.norm(x_hat)

    # construct transform
    transform = np.c_[x_hat, y_hat, z_hat, pos]
    transform = np.r_[transform, np.array([0, 0, 0, 1]).reshape(1, 4)]

    # if the determinant is 0 then this is a reflection, to undo that the direciton of the
    # epicondylar axis should be switched

    if np.round(np.linalg.det(transform)) == -1:
        transform[:, 0] *= -1

    # return transform for CT csys -> canal-epi csys
    transform = utils.inv_transform(transform)
    return transform
